#include <msp/core/refptr.h>
#include <msp/net/resolve.h>
#include <msp/time/units.h>
#include "client.h"
#include "request.h"
#include "response.h"

using namespace std;

namespace Msp {
namespace Http {

Client::Client():
	sock(0),
	event_disp(0),
	resolver(0),
	resolve_listener(0),
	resolve_tag(0),
	user_agent("libmsphttp/0.1"),
	request(0),
	response(0)
{ }

Client::~Client()
{
	delete sock;
	delete request;
	delete response;
}

void Client::use_event_dispatcher(IO::EventDispatcher *ed)
{
	if(event_disp && sock)
		event_disp->remove(*sock);
	event_disp = ed;
	if(event_disp && sock)
		event_disp->add(*sock);
}

void Client::use_resolver(Net::Resolver *r)
{
	if(resolver)
	{
		delete resolve_listener;
		resolve_listener = 0;
	}

	resolver = r;
	if(resolver)
		resolve_listener = new ResolveListener(*this);
}

void Client::start_request(const Request &r)
{
	if(request)
		throw client_busy();

	delete sock;
	sock = 0;

	request = new Request(r);
	if(!user_agent.empty())
		request->set_header("User-Agent", user_agent);

	delete response;
	response = 0;
	in_buf.clear();

	string host = r.get_header("Host");
	if(host.find(':')==string::npos)
		host += ":80";
	if(resolver)
		resolve_tag = resolver->resolve(host);
	else
	{
		RefPtr<Net::SockAddr> addr = Net::resolve(host);
		address_resolved(resolve_tag, *addr);
	}
}

const Response *Client::get_url(const std::string &url)
{
	start_request(Request::from_url(url));
	wait_response();
	return response;
}

void Client::tick()
{
	if(!request)
		return;

	while(IO::PollEvent ev = IO::poll(*sock, sock->get_events(), Time::zero))
		sock->event(ev);

	if(response && response->is_complete())
	{
		signal_response_complete.emit(*response);

		delete sock;
		sock = 0;
		delete request;
		request = 0;
	}
}

void Client::wait_response()
{
	while(request && (!response || !response->is_complete()))
		tick();
}

void Client::abort()
{
	delete sock;
	sock = 0;
	delete request;
	request = 0;
}

void Client::address_resolved(unsigned tag, const Net::SockAddr &addr)
{
	if(tag!=resolve_tag)
		return;
	resolve_tag = 0;

	sock = new Net::StreamSocket(addr.get_family());
	sock->set_block(false);

	sock->signal_data_available.connect(sigc::mem_fun(this, &Client::data_available));
	sock->signal_connect_finished.connect(sigc::mem_fun(this, &Client::connect_finished));
	if(event_disp)
		event_disp->add(*sock);

	sock->connect(addr);
}

void Client::resolve_failed(unsigned tag, const exception &err)
{
	if(tag!=resolve_tag)
		return;
	resolve_tag = 0;

	signal_socket_error.emit(err);

	delete request;
	request = 0;
}

void Client::connect_finished(const exception *err)
{
	if(err)
	{
		signal_socket_error.emit(*err);

		delete request;
		request = 0;
	}
	else
		sock->write(request->str());
}

void Client::data_available()
{
	char rbuf[4096];
	unsigned len;
	try
	{
		len = sock->read(rbuf, sizeof(rbuf));
	}
	catch(const exception &e)
	{
		signal_socket_error.emit(e);
		return;
	}

	if(!len)
		return;
	in_buf.append(rbuf, len);

	if(!response)
	{
		if(in_buf.find("\r\n\r\n")!=string::npos || in_buf.find("\n\n")!=string::npos)
		{
			response = new Response(Response::parse(in_buf));
			response->set_user_data(request->get_user_data());
			in_buf = string();
		}
	}
	else
	{
		len = response->parse_content(in_buf);
		in_buf.erase(0, len);
	}

	if(response && response->is_complete())
	{
		signal_response_complete.emit(*response);

		delete request;
		request = 0;
	}
}


Client::ResolveListener::ResolveListener(Client &c):
	client(c)
{
	client.resolver->signal_address_resolved.connect(sigc::mem_fun(this, &ResolveListener::address_resolved));
	client.resolver->signal_resolve_failed.connect(sigc::mem_fun(this, &ResolveListener::resolve_failed));
}

void Client::ResolveListener::address_resolved(unsigned tag, const Net::SockAddr &addr)
{
	client.address_resolved(tag, addr);
}

void Client::ResolveListener::resolve_failed(unsigned tag, const exception &err)
{
	client.resolve_failed(tag, err);
}

} // namespace Http
} // namespace Msp
